-- Conway's Game Of Life Simulator
-- Written by Inny
-- Please read the Licensing Terms at the end of this file.

-- This file has been extensively commented for the purpose of use as an
-- educational tool.

--------------------------------------------------------------------------------

-- Variables are created via "Assignment" using the equals symbol.
-- This variable, game_name, contains a piece of text called a "string"
-- and lives at the "Global scope", meaning everything will be able
-- to access it.
game_name = "Inny's Game of Life.love"

-- This is another way to write strings in Lua, using [[ and ]] instead of
-- double quotes.
help_text = [[
This is a "Conway's Game Of Life" simulator.

Each cell on the screen is being updated based on automata rules and conditions of the frame before.

If the cell is "Alive", it'll "Die" on the next frame when the surrounding 8 cells are less than 2, or greater than 3.

If a cell is "Dead", it'll "Reproduce" if the number of surrounding cells are exactly 3.

-- Keyboard Controls --
H: Show or hide this info display.
Space: Pause or Unpause the simulation.
C: Clear the world.
F: Fill the world.
1-9: Randomize the world, with 10% to 90% density.
F5: Change window size.
F10: Quit.

-- Mouse Controls --
Left Click: Change a cell state.
Left Drag: Draw a state across multiple cells.

Try drawing a few of these lifeforms!
]]

-- The local keyword creates a variable that is "Scoped" which means
-- that it can only be accessed in a context. Lua gives these variables
-- special meaning and quicker access. lg is being assigned a table that
-- already exists, and this is normally known as "Aliasing". 
local lg = love.graphics

-- Tables are a data structure that allow you to keep other variables
-- inside of it. This keeps your code cleaner, and makes it easier to
-- pass data between functions. As shown above, the dot between the table
-- name and the variable name is how we indicate which variable in the
-- table is the one we want to access.
color = {
  darker_gray = {  64,  64,  64 }, -- Red, Green, Blue
  dark_gray   = {  80,  80,  80 },
  light_gray  = { 192, 192, 192 },
  white       = { 255, 255, 255 },
  info_border = { 128, 128, 128 }, 
  info_back   = {  64,  64,  64, 240 }, -- Red, Green, Blue, Alpha
}

-- The variable names in a table are really just other values. In
-- this case, I'm using strings that have numbers in them. Those
-- strings are different from actual numbers. The square brackets
-- are necessary to do it this way.
randomize_key_density = {
  ['1'] = 10, ['2'] = 20, ['3'] = 30, ['4'] = 40, ['5'] = 50,
  ['6'] = 60, ['7'] = 70, ['8'] = 80, ['9'] = 90,
}

-- These strings will be used for drawing some sample lifeforms.
-- When you see \n, this is how strings keep track of where lines end
-- and a new one begins. Here I had to write them, but in the larger
-- string above (help_text) I didn't have to write them because there
-- are actually new lines there.
life_forms = {
  ["Glider"] = " # \n  #\n###",
  ["Blinker"] = "   \n###\n   ",
  ["Toad"] = "  # \n#  #\n#  #\n #  ",
  ["Block"] = "##\n##",
}

-- Some tables don't require that you list the names of the variables
-- it has inside. When written in this form, the variable names are
-- implied to be the numbers 1, 2, 3, and so on (and not strings).
-- This kind of table is known as an "Array".
window_size_list = {
  { 480, 480 },
  { 640, 640 },
  { 800, 800 },
}
-- These are other global variables.
clock = 0
paused = false -- false is a "boolean" value
info_display = true -- true is also boolean, and different from false
click_type = false

-- These two variables contain values are the result of a function call.
-- The parenthesis is how we make the function call, and the numbers
-- contained inside are used as parameters.
small_font = lg.newFont(12)
big_font = lg.newFont(24)

--------------------------------------------------------------------------------
-- Functions are how pieces of code are saved for reuse. By putting code
-- inside a function, it won't be executed right now. It'll be executed
-- later when you use the function's name in a statement. The variable
-- names between the parenthesis are "Paramaters" and can be different
-- each time the function runs, so that it does something different each
-- time. Functions can also return values.

-- Inside of a function is a "context", and local keywords are used
-- to keep variables only inside the function.
function drawLifeForm(name, x, y)
  -- block is being used to copy from the lifeforms table, which were strings.
  -- We use the square brackets
  local block = life_forms[name]

  -- these variables track the size and location of what we're drawing
  local width = 0
  local tx, ty = x, y+16

  -- A "For Loop" will execute a block of code multiple times.
  -- i is a general name for the "incrementor", which is what will
  -- change each time the loop runs. It'll run between 1 and "#block",
  -- which is the length of the string.
  for i = 1, #block do

    -- char is an individual character from the text in that string.
    local char = block:sub(i, i)

    -- An if statement is how you compare variables and values. When
    -- variables are equal to the values you expect, it changes which
    -- code runs. For instance, when char is a newline character, then
    -- only the code in the first block runs.
    if char == '\n' then
      width = math.max(width, tx-x)  -- notice the table and function call here.
      tx = x
      ty = ty + 8

    -- When char is a space character, then the first block didn't run,
    -- but the second block will. "elseif" checks a second condition.
    elseif char == ' ' then
      tx = tx + 8

    -- And a third condition is checked, because we're using the three
    -- characters to decide what to draw and what not to.
    elseif char == '#' then
      lg.rectangle("fill", tx, ty, 7, 7)
      tx = tx + 8
    end
  end

  -- And the end, we write the name of the Lifeform above it.
  lg.printf(name, x, y, width, "center")
end

-- This is an alternate way to write functions. What's interesting
-- about it is that the function's name is a regular variable, and
-- the function is behaving like a value. This is called a
-- "First Class Function" and all lua functions are this way.
changeWindowSize = function()
  local scr_width, scr_height = lg.getWidth(), lg.getHeight()

  -- the ipairs function is how you work your way through a table
  -- that uses numbers, which we called an "Array", to check each
  -- slot or "index" of that array. In this form of the For loop,
  -- index and size are assigned the number and the subtable.
  for index, size in ipairs(window_size_list) do
    if size[1]==scr_width and size[2]==scr_height then
      local next_index = (index % #window_size_list) + 1
      local new_size = window_size_list[next_index]
      lg.setMode(new_size[1], new_size[2])

      -- game_world will be explained later.
      game_world = World.new(math.floor(new_size[1]/8), math.floor(new_size[2]/8))
      game_world:randomize(20)

      break -- We can end a loop early with this keyword
    end
  end
end

--------------------------------------------------------------------------------
-- As mentioned before, tables are useful for storing functions in.
-- We've seen this with the lg and math tables. However, what's also
-- useful is "Object Oriented" programming, where a table of functions
-- are used as a "Class" and other tables can use that class to
-- have the same functions, without needing to explicitly copy every
-- function each time we want to. In lua, this is accomplished with
-- "Metatables", that instruct tables on where to find the functions
-- that they want. To start a class definition, we make a regular table.

World = {}

-- We can add new variables to tables at anytime using the dot symbol.
-- the __index variable tells all tables using the World metatable that
-- the functions that they're looking for are in the World table as well.
World.__index = World  

-- For convience, we'll use the "new" function as a way to easily
-- make tables that use the World table as their class.
function World.new(width, height)
  local self = {
    width = width,
    height = height
  }
  -- The setmetatable function is how we connect a table and a metatable.
  setmetatable(self, World)

  self.data = {}
  self.last = {}
  for i = 1, self.width * self.height do
    self.data[i] = false 
    self.last[i] = false
  end

  -- The return keyword exits a function, and the variables or values
  -- that you list on the line are things that carry to outside the function.
  -- When you see a function call on the right side of an equal side, or
  -- inside a parenthesis, this is where functions connect together.
  return self
end

-- Lua uses the colon symbol to make object oriented programming easier.
-- What it does is remove the need to write a parameter named self
-- from the function's parameter list. This is called "Sugar Syntax."
-- The next four functions are used to make other functions cleaner to read.
function World:pos(x, y)
  local w = self.width
  return 1+ ((x-1)%w) + (((y-1)%self.height)*w)
end

function World:set(x, y, alive)
  self.data[self:pos(x, y)] = alive
end

function World:get(x, y)
  return self.data[self:pos(x, y)]
end

function World:getLast(x, y)
  return self.last[self:pos(x, y)]
end

-- As mentioned above, the randomize function is now easier to read
-- because the set function exists. This kind of programming is known
-- as "Abstraction," which means that we can define terms that make
-- in easier to understand what other code does.
function World:randomize(density)
  for y = 1, self.width do
    for x = 1, self.height do
      local random_value = math.random(1, 100)
      -- comparisons can happen anywhere, not just in if statements.
      -- When a comparison happens, it either is the value "true" or
      -- the value "false", which gets stored in the variable alive.
      local alive = (random_value <= density) 
      self:set(x, y, alive)
    end
  end
end

-- I've been leaving some of these functions uncommented so that you can
-- try to figure out on your own what they're doing.
function World:fill(alive)
  for y = 1, self.width do
    for x = 1, self.height do
      self:set(x, y, alive)
    end
  end
end

function World:step()
  self.data, self.last = self.last, self.data
  for y = 1, self.width do
    for x = 1, self.height do
      -- Have you seen the applyRules function yet? How can we use it
      -- if we don't know what it is? Well, because this code doesn't
      -- run until we use it later, that means that we use abstractions
      -- before we've defined them.
      self:applyRules(x, y) 
    end
  end
end

-- The 'do' keyword creates a contextual block. It continues to execute
-- as normal, but when a local keyword is encountered, the variables
-- are hidden in the block, and things outside the block can't access them.
do
  -- nlist is only visible to the countLastNeighbors function.
  local nlist = {
    {-1, -1}, { 0, -1}, {1, -1}, {-1, 0},
    { 1,  0}, {-1,  1}, {0,  1}, { 1, 1}
  }

  function World:countLastNeighbors(x, y)
    local neighbors = 0
    for _, loc in ipairs(nlist) do
      local alive = self:getLast(x+loc[1], y+loc[2])
      if alive then neighbors = neighbors + 1 end
    end
    return neighbors
  end
end

-- Here's the applyRules function! Because the chain of execution
-- doesn't execute any functions until they're actually used,
-- you're free to define functions where ever you want, as long
-- as it's before when it gets used. Do you remember where applyRules
-- was used? It's usually a good idea to keep functions that use each
-- other where they're being used.

function World:applyRules(x, y)
  -- As this is Conway's Game Of Life, this function is where
  -- the primary rules of the game are coded.
  local alive = self:getLast(x, y)
  local neighbors = self:countLastNeighbors(x, y)

  -- If statements can get a bit complicated, so we can use
  -- parenthesis to make it clearer what we mean. In this case we check
  -- if it's alive first, before the rest. When alive, then we check
  -- the other two conditions in the parenthesis.
  if alive and ((neighbors < 2) or (neighbors > 3)) then
    alive = false

  -- Parenthesis aren't always necessary, but you'll need to know the
  -- "Order Of Operations", which is how Lua decides what you meant.
  elseif (not alive) and (neighbors == 3) then
    alive = true
  end

  self:set(x, y, alive)
end


--------------------------------------------------------------------------------
-- The Love Framework requires that you define "Handlers", which
-- are functions that Love uses to play your game. You'll see
-- functions being defined here in the love table. That's where they
-- want us to put them.

function love.load()
  -- Here we're not using a local keyword, because we're instead
  -- putting a value in the "world" global variable. Lua allows
  -- you to define global variables anywhere, simply by leaving off
  -- the local keyword. Be careful when you do this, because sometimes
  -- it might not be clear where global variables are defined and
  -- used. Do you remember where else game_world was used before?
  game_world = World.new(80, 80)

  -- Now that game_world exists, and the "new" function made it an object
  -- with World as its class, all of World's functions can be used by
  -- game_world. When you use the colon syntax with objects, then the
  -- "self" variable in functions also using the colon will be filled
  -- with the object in use.
  game_world:randomize(20)
end

-- The draw function is used by love to draw what the screen looks like.
-- This function is typically called about 60 times in a second, though
-- it can be a lot faster or a lot slower depending on how fast your
-- computer is and how fast its screen wants to be redrawn.
function love.draw()
  local scr_width, scr_height = lg.getWidth(), lg.getHeight()
  lg.setColor(paused and color.darker_gray or color.dark_gray)
  for y = 0, scr_height, 8 do
    lg.line(0, y, scr_width, y)
  end
  for x = 0, scr_width, 8 do
    lg.line(x, 0, x, scr_height)
  end

  lg.setColor(paused and color.light_gray or color.white)
  for y = 1, game_world.height do
    for x = 1, game_world.width do
      if game_world:get(x, y) then
        lg.rectangle("fill", (x-1)*8+1, (y-1)*8+1, 5, 5)
      end
    end
  end

  if info_display then
    -- because the window can change size, I can't used any regular
    -- numbers to put the help display in the center of the screen,
    -- so I use these variables to figure out where that center
    -- will be, and the top left of the box that goes there.
    local base_x, base_y = (scr_width/2)-240, (scr_height/2)-240

    -- this draws the box of the help display
    lg.setColor(color.info_back)
    lg.rectangle("fill", base_x, base_y, 480, 480)
    lg.setColor(color.info_border)
    lg.rectangle("line", base_x, base_y, 480, 480)

    -- this draws the bigger text in that help display
    lg.setColor(color.white)
    lg.setFont(big_font)
    lg.printf(game_name, base_x, base_y+20, 480, "center")

    -- this draws the smaller text with all of the rules and controls in it.
    lg.setFont(small_font)
    lg.printf(help_text, base_x+40, base_y+60, 400)
    
    -- You should remember where the drawLifeForm function was first
    -- introduced. We're using that function here to draw the lifeforms
    -- in the help window.
    drawLifeForm("Glider", base_x+80, base_y+420)
    drawLifeForm("Blinker", base_x+180, base_y+420)
    drawLifeForm("Toad", base_x+280, base_y+420)
    drawLifeForm("Block", base_x+380, base_y+420)
  end
end

-- The update function is called just as often as the draw function is.
-- This is where Love wants you to actually change everything. Because
-- my simulation runs at 10hz, I need to keep track of time, so that
-- I only run the simulation every 6 or so times update is called. To
-- do that, I use the "dt" parameter, which tells me how much time
-- passed since the last time it was called. Instead of counting 6
-- function calls, I count 0.1 seconds to get a more smooth and
-- accurate speed.
function love.update(dt)

  -- Love doesn't tell us when the mouse is being moved, we have to
  -- ask with the getPosition, and I only care about that when you're
  -- holding down the left mouse button.
  if love.mouse.isDown('l') then
    x, y = love.mouse.getPosition()
    x = math.floor(x / 8) + 1
    y = math.floor(y / 8) + 1
    game_world:set(x, y, click_type)
  end

  if not paused then
    clock = clock + dt
    -- There's a secret keyboard control here. If you hold the backquote
    -- key, then the simulation will run as fast as your computer will
    -- allow.
    if (clock >= 0.1) or love.keyboard.isDown('`') then
      game_world:step()

      -- The 0.2 listed here is a kind of safety measure. If for some
      -- reason the computer hung up for a few seconds, when it comes
      -- back to life I want it to forget about how long it was
      -- hung up for. This is to prevent a second hangup, or a sudden
      -- radical speedup out of nowhere.
      if (clock >= 0.1) and (clock <= 0.2) then
        clock = clock - 0.1
      else
        clock = 0
      end
    end
  end
end

-- the keypressed function is where we tell love what to do when
-- a key has been pressed on the keyboard. This is where all of the
-- game's controls are coded. You'll notice that a lot of global
-- variables are being used here.
function love.keypressed(k)
  if k == ' ' then
    paused = not paused
  elseif k == 'c' then
    game_world:fill(false)
  elseif k == 'f' then
    game_world:fill(true)
  elseif k == 'h' then
    info_display = not info_display
  elseif k == "f5" then
    changeWindowSize()
  elseif k == "f10" then
    love.event.quit()
  elseif randomize_key_density[k] then
    game_world:randomize(randomize_key_density[k])
  end
end

-- Love does react to the mouse being pressed. When it happens,
-- I want to know what was under the mouse at that moment, so that
-- we can draw the correct state for all of the game cells as
-- the cursor drags.
function love.mousepressed(x, y, b)
  if b == 'l' then
    x = math.floor(x / 8) + 1
    y = math.floor(y / 8) + 1
    click_type = not game_world:get(x, y)
    game_world:set(x, y, click_type)
  end
end

--------------------------------------------------------------------------------
-- If Lua and Love have interested you, please read more about them!

-- Love is an open source framework developed at http://love2d.org
-- They have a very friendly Forum and community.

-- The Book "Programming In Lua" is a highly recommended read.
-- The first edition is freely availiable online.
-- See: http://www.lua.org/pil/

-- The Lua Reference Manual is also freely available online, however
-- the contents are meant for advanced programmers. It's recommended
-- that you try to read it, but don't be discouraged if you don't
-- understand everything that you read in it.
-- See: http://www.lua.org/manual/
--------------------------------------------------------------------------------
--[[

This file is Licensed under the terms of the ZLIB license:

Copyright (c) 2013 William C. Bubel

This software is provided 'as-is', without any express or implied warranty. In
no event will the authors be held liable for any damages arising from the use
of this software.

Permission is granted to anyone to use this software for any purpose, including
commercial applications, and to alter it and redistribute it freely, subject to
the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in a
    product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.

--]]

